/****************************************************************************
** $Id: dl_dxf.cpp,v 1.30 2003/11/27 23:45:17 andrew Exp $
**
** Copyright (C) 2001-2003 RibbonSoft. All rights reserved.
**
** This file is part of the dxflib project.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding valid dxflib Professional Edition licenses may use 
** this file in accordance with the dxflib Commercial License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.ribbonsoft.com for further details.
**
** Contact info@ribbonsoft.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include "dl_dxf.h"

#include <algorithm>
#include <string>
#include <sstream>
#include <cstdio>
#include <cassert>
#include <cmath>

#include "dl_attributes.h"
#include "dl_codes.h"
#include "dl_creationinterface.h"
#include "dl_writer_ascii.h"

#if defined(__OS2__)||defined(__EMX__)||defined(WIN32)
#define strcasecmp(s,t) stricmp(s,t)
#endif


/**
 * Default constructor.
 */
DL_Dxf::DL_Dxf() {
    styleHandleStd = 0;
    version = VER_2000;

    vertices = NULL;
    maxVertices = 0;
    vertexIndex = 0;

    leaderVertices = NULL;
    maxLeaderVertices = 0;
    leaderVertexIndex = 0;

    hatchLoops = NULL;
    maxHatchLoops = 0;
    hatchLoopIndex = -1;
    hatchEdges = NULL;
    maxHatchEdges = NULL;
    hatchEdgeIndex = NULL;
	dropEdges = false;

    bulge = 0.0;
}



/**
 * Destructor.
 */
DL_Dxf::~DL_Dxf() {
    if (vertices!=NULL) {
        delete[] vertices;
    }
    if (leaderVertices!=NULL) {
        delete[] leaderVertices;
    }
    if (hatchLoops!=NULL) {
        delete[] hatchLoops;
    }
    if (hatchEdges!=NULL) {
        for (int i=0; i<maxHatchLoops; ++i) {
			if (hatchEdges[i]!=NULL) {
            	delete[] hatchEdges[i];
			}
        }
        delete[] hatchEdges;
    }
    if (maxHatchEdges!=NULL) {
        delete[] maxHatchEdges;
    }
    if (hatchEdgeIndex!=NULL) {
        delete[] hatchEdgeIndex;
    }
}



/**
 * @brief Reads the given file and calls the appropriate functions in
 * the given creation interface for every entity found in the file.
 *
 * @param file Input
 *		Path and name of file to read
 * @param creationInterface
 *		Pointer to the class which takes care of the entities in the file.
 *
 * @retval true If \p file could be opened.
 * @retval false If \p file could not be opened.
 */
bool DL_Dxf::in(const string& file, DL_CreationInterface* creationInterface) {
    FILE *fp;
    firstCall = true;
    currentEntity = DL_UNKNOWN;

    fp = fopen(file.c_str(), "rt");
    if (fp) {
        while (readDxfGroups(fp, creationInterface)) {}
        fclose(fp);
        return true;
    }

    return false;
}



/**
 * Reads a DXF file from an existing stream.
 *
 * @param stream The string stream.
 * @param creationInterface
 *		Pointer to the class which takes care of the entities in the file.
 *
 * @retval true If \p file could be opened.
 * @retval false If \p file could not be opened.
 */
#ifndef __GCC2x__
bool DL_Dxf::in(std::stringstream& stream,
                DL_CreationInterface* creationInterface) {

    if (stream.good()) {
        firstCall=true;
        currentEntity = DL_UNKNOWN;
        while (readDxfGroups(stream, creationInterface)) {}
        return true;
    }
    return false;
}
#endif



/**
 * @brief Reads a group couplet from a DXF file.  Calls another function
 * to process it.
 *
 * A group couplet consists of two lines that represent a single
 * piece of data.  An integer constant on the first line indicates
 * the type of data.  The value is on the next line.\n
 *
 * This function reads a couplet, determines the type of data, and
 * passes the value to the the appropriate handler function of
 * \p creationInterface.\n
 * 
 * \p fp is advanced so that the next call to \p readDXFGroups() reads
 * the next couplet in the file.
 *
 * @param fp Handle of input file
 * @param creationInterface Handle of class which processes entities
 *		in the file
 *
 * @retval true If EOF not reached.
 * @retval false If EOF reached.
 */
bool DL_Dxf::readDxfGroups(FILE *fp, DL_CreationInterface* creationInterface) {

    // Read one group of the DXF file and chop the lines:
    if (DL_Dxf::getChoppedLine(groupCodeTmp, DL_DXF_MAXLINE, fp) &&
            DL_Dxf::getChoppedLine(groupValue, DL_DXF_MAXLINE, fp) ) {

        groupCode = (unsigned int)atoi(groupCodeTmp);

        //std::cout << groupCode << "\n";
        //std::cout << groupValue << "\n";

        processDXFGroup(creationInterface, groupCode, groupValue);
    }

    return !feof(fp);
}



/**
 * Same as above but for stringstreams.
 */
#ifndef __GCC2x__
bool DL_Dxf::readDxfGroups(std::stringstream& stream,
                           DL_CreationInterface* creationInterface) {

    // Read one group of the DXF file and chop the lines:
    if (DL_Dxf::getChoppedLine(groupCodeTmp, DL_DXF_MAXLINE, stream) &&
            DL_Dxf::getChoppedLine(groupValue, DL_DXF_MAXLINE, stream) ) {

        groupCode = (unsigned int)atoi(groupCodeTmp);

        processDXFGroup(creationInterface, groupCode, groupValue);
    }
    return !stream.eof();
}
#endif



/**
 * @brief Reads line from file & strips whitespace at start and newline 
 * at end.
 *
 * @param s Output\n
 *		Pointer to character array that chopped line will be returned in.
 * @param size Size of \p s.  (Including space for NULL.)
 * @param fp Input\n
 *		Handle of input file.
 *
 * @retval true if line could be read
 * @retval false if \p fp is already at end of file
 *
 * @todo Change function to use safer FreeBSD strl* functions
 * @todo Is it a problem if line is blank (i.e., newline only)?
 *		Then, when function returns, (s==NULL).
 */
bool DL_Dxf::getChoppedLine(char *s, unsigned int size, FILE *fp) {
    if (!feof(fp)) {
        // The whole line in the file.  Includes space for NULL.
        char* wholeLine = new char[size];
        // Only the useful part of the line
        char* line;

        line = fgets(wholeLine, size, fp);

        if (line!=NULL && strlen(line)!=0) { // Evaluates to fgets() retval
            // line == wholeLine at this point.
            // Both guaranteed to be NULL terminated.

            // Strip leading whitespace and trailing CR/LF.
            stripWhiteSpace(line);

            strncpy(s, line, size);
            s[size] = '\0';
            // s should always be NULL terminated, because:
            assert(size > strlen(line));
        }

        delete[] wholeLine; // Done with wholeLine

        return true;
    } else {
        s[0] = '\0';
        return false;
    }
}



/**
 * Same as above but for stringstreams.
 */
#ifndef __GCC2x__
bool DL_Dxf::getChoppedLine(char *s, unsigned int size,
                            std::stringstream& stream) {

    if (!stream.eof()) {
        // The whole line in the file.  Includes space for NULL.
        char* wholeLine = new char[size];
        // Only the useful part of the line
        stream.getline(wholeLine, size);
        stripWhiteSpace(wholeLine);
        strncpy(s, wholeLine, size);
        assert(size > strlen(wholeLine));
        delete[] wholeLine; // Done with wholeLine
        return true;
    } else {
        s[0] = '\0';
        return false;
    }
}
#endif



/**
 * @brief Strips leading whitespace and trailing Carriage Return (CR)
 * and Line Feed (LF) from NULL terminated string.
 *
 * @param s Input and output.
 *		NULL terminates string.
 *
 * @retval true if \p s is non-NULL
 * @retval false if \p s is NULL
 */
bool DL_Dxf::stripWhiteSpace(char *s) {
    // Skip whitespace, excluding \n, at beginning of line
    while (s[0]==' ' || s[0]=='\t') {
        ++s;
    }

    int lastChar = strlen(s) - 1; // Last non-NULL character in s

    // Is last character CR or LF?
    while ( (lastChar >= 0) &&
            ((s[lastChar] == '\n') || (s[lastChar] == '\r')) ) {
        // Replace CR or LF with NULL and decrement string index
        s[lastChar--] = '\0';
    }

    return ((s) ? true : false);
}



/**
 * Processes a group (pair of group code and value).
 *
 * @param creationInterface Handle to class that creates entities and
 * other CAD data from DXF group codes
 *
 * @param groupCode Constant indicating the data type of the group.
 * @param groupValue The data value.
 *
 * @retval true if done processing current entity and new entity begun
 * @retval false if not done processing current entity
*/
bool DL_Dxf::processDXFGroup(DL_CreationInterface* creationInterface,
                             int groupCode, const char *groupValue) {



    // Init on first call
    if (firstCall) {
        for (int i=0; i<DL_DXF_MAXGROUPCODE; ++i) {
            values[i][0] = '\0';
        }
        settingValue[0] = '\0';
        firstCall=false;
    }

    // Indicates start of new entity or var
    if (groupCode==0 || groupCode==9) {

        // If new entity is encountered, the last one must be complete
        // prepare attributes which can be used for most entities:
        char name[DL_DXF_MAXLINE+1];
        if (strlen(values[8])!=0) {
            strcpy(name, values[8]);
        }
        // defaults to layer '0':
        else {
            strcpy(name, "0");
        }

        int width;
        // Compatibillity with qcad1:
        if (strlen(values[39])!=0 &&
                strlen(values[370])==0) {
            width = toInt(values[39], -1);
        }
        // since autocad 2002:
        else if (strlen(values[370])!=0) {
            width = toInt(values[370], -1);
        }
        // default to BYLAYER:
        else {
            width = -1;
        }

        int color;
        color = toInt(values[62], 256);

        char linetype[DL_DXF_MAXLINE+1];
        strcpy(linetype, toString(values[6], "BYLAYER"));

        attrib = DL_Attributes(values[8],          // layer
                               color,              // color
                               width,              // width
                               linetype);          // linetype
        creationInterface->setAttributes(attrib);


        // Add the last entity via creationInterface
        switch (currentEntity) {
        case DL_SETTING:
            for (int c=0; c<=380; ++c) {
                if (values[c][0]!='\0') {
                    // string
                    if (c>=0 && c<=9) {
                        creationInterface->setVariableString(settingKey,
                                                             values[c], c);
                    }
                    // double
                    else if (c>=10 && c<=39) {
                        if (c==10) {
                            creationInterface->setVariableVector(
                                settingKey,
                                toReal(values[c]),
                                toReal(values[c+10]),
                                toReal(values[c+20]),
                                c);
                        }
                    }
                    // double
                    else if (c>=40 && c<=59) {
                        creationInterface->setVariableDouble(settingKey,
                                                             toReal(values[c]),
                                                             c);
                    }
                    // int
                    else if (c>=60 && c<=99) {
                        creationInterface->setVariableInt(settingKey,
                                                          toInt(values[c]),
                                                          c);
                    }
                    // misc
                    else {
                        creationInterface->setVariableString(settingKey,
                                                             values[c],
                                                             c);
                    }
                }
            }
            break;

        case DL_LAYER: {
                // correct some impossible attributes for layers:
                attrib = creationInterface->getAttributes();
                if (attrib.getColor()==256 || attrib.getColor()==0) {
                    attrib.setColor(7);
                }
                if (attrib.getWidth()<0) {
                    attrib.setWidth(1);
                }
                if (!strcasecmp(attrib.getLineType().c_str(), "BYLAYER") ||
                        !strcasecmp(attrib.getLineType().c_str(), "BYBLOCK")) {
                    attrib.setLineType("CONTINUOUS");
                }

                // add layer
                creationInterface->addLayer(DL_LayerData(values[2],
                                            toInt(values[70])));
            }
            break;

        case DL_BLOCK: {
                //std::cout << "dxflib: adding block: \"" << values[2] << "\"\n";
                DL_BlockData d(
                    // Name:
                    values[2],
                    // flags:
                    toInt(values[70]),
                    // base point:
                    toReal(values[10]),
                    toReal(values[20]),
                    toReal(values[30]));

                creationInterface->addBlock(d);
            }
            break;

        case DL_ENDBLK: {
                creationInterface->endBlock();
            }
            break;

        case DL_ENTITY_POINT: {
                DL_PointData d(toReal(values[10]),
                               toReal(values[20]),
                               toReal(values[30]));
                creationInterface->addPoint(d);
            }
            break;

        case DL_ENTITY_LINE: {
                DL_LineData d(toReal(values[10]),
                              toReal(values[20]),
                              toReal(values[30]),
                              toReal(values[11]),
                              toReal(values[21]),
                              toReal(values[31]));

                creationInterface->addLine(d);
            }
            break;

        case DL_ENTITY_POLYLINE:
            bulge = toReal(values[42]);
            // fall through
        case DL_ENTITY_LWPOLYLINE: {
                DL_PolylineData pd(maxVertices, toInt(values[70], 0));
                creationInterface->addPolyline(pd);

                if (currentEntity==DL_ENTITY_LWPOLYLINE) {
                    for (int i=0; i<maxVertices; i++) {
                        DL_VertexData d(vertices[i*4],
                                        vertices[i*4+1],
                                        vertices[i*4+2],
                                        vertices[i*4+3]);

                        creationInterface->addVertex(d);
                    }
                }
            }
            break;

        case DL_ENTITY_VERTEX: {
                DL_VertexData d(toReal(values[10]),
                                toReal(values[20]),
                                toReal(values[30]),
                                bulge);
                //toReal(values[42]));

                bulge = toReal(values[42]);

                creationInterface->addVertex(d);
            }
            break;

        case DL_ENTITY_ARC: {
                DL_ArcData d(toReal(values[10]),
                             toReal(values[20]),
                             toReal(values[30]),
                             toReal(values[40]),
                             toReal(values[50]),
                             toReal(values[51]));

                creationInterface->addArc(d);
            }
            break;

        case DL_ENTITY_CIRCLE: {
                DL_CircleData d(toReal(values[10]),
                                toReal(values[20]),
                                toReal(values[30]),
                                toReal(values[40]));

                creationInterface->addCircle(d);
            }
            break;

        case DL_ENTITY_ELLIPSE: {
                DL_EllipseData d(toReal(values[10]),
                                 toReal(values[20]),
                                 toReal(values[30]),
                                 toReal(values[11]),
                                 toReal(values[21]),
                                 toReal(values[31]),
                                 toReal(values[40], 1.0),
                                 toReal(values[41], 0.0),
                                 toReal(values[42], 2*M_PI));

                creationInterface->addEllipse(d);
            }
            break;

        case DL_ENTITY_INSERT: {
                DL_InsertData d(values[2],
                                // insertion point
                                toReal(values[10], 0.0),
                                toReal(values[20], 0.0),
                                toReal(values[30], 0.0),
                                // scale:
                                toReal(values[41], 1.0),
                                toReal(values[42], 1.0),
                                toReal(values[43], 1.0),
                                // angle:
                                toReal(values[50], 0.0),
                                // cols / rows:
                                toInt(values[70], 1),
                                toInt(values[71], 1),
                                // spacing:
                                toReal(values[44], 0.0),
                                toReal(values[45], 0.0));

                creationInterface->addInsert(d);
            }
            break;

        case DL_ENTITY_MTEXT: {
                double angle = 0.0;

                if (values[50][0]!='\0') {
                    angle = toReal(values[50], 0.0);
                } else if (values[11][0]!='\0' && values[21][0]!='\0') {
                    double x = toReal(values[11], 0.0);
                    double y = toReal(values[21], 0.0);

                    if (fabs(x)<1.0e-6) {
                        if (y>0.0) {
                            angle = M_PI/2.0;
                        } else {
                            angle = M_PI/2.0*3.0;
                        }
                    } else {
                        angle = atan(y/x);
                    }
                }

                DL_MTextData d(
                    // insertion point
                    toReal(values[10], 0.0),
                    toReal(values[20], 0.0),
                    toReal(values[30], 0.0),
                    // height
                    toReal(values[40], 2.5),
                    // width
                    toReal(values[41], 100.0),
                    // attachment point
                    toInt(values[71], 1),
                    // drawing direction
                    toInt(values[72], 1),
                    // line spacing style
                    toInt(values[73], 1),
                    // line spacing factor
                    toReal(values[44], 1.0),
                    // text
                    values[1],
                    // style
                    values[7],
                    // angle
                    angle);
                creationInterface->addMText(d);
            }
            break;

        case DL_ENTITY_TEXT: {
                DL_TextData d(
                    // insertion point
                    toReal(values[10], 0.0),
                    toReal(values[20], 0.0),
                    toReal(values[30], 0.0),
                    // alignment point
                    toReal(values[11], 0.0),
                    toReal(values[21], 0.0),
                    toReal(values[31], 0.0),
                    // height
                    toReal(values[40], 2.5),
                    // x scale
                    toReal(values[41], 1.0),
                    // generation flags
                    toInt(values[71], 0),
                    // h just
                    toInt(values[72], 0),
                    // v just
                    toInt(values[73], 0),
                    // text
                    values[1],
                    // style
                    values[7],
                    // angle
                    (toReal(values[50], 0.0)*2*M_PI)/360.0);

                creationInterface->addText(d);
            }
            break;
			
        case DL_ENTITY_ATTRIB: {
                DL_TextData d(
                    // insertion point
                    toReal(values[10], 0.0),
                    toReal(values[20], 0.0),
                    toReal(values[30], 0.0),
                    // alignment point
                    toReal(values[11], 0.0),
                    toReal(values[21], 0.0),
                    toReal(values[31], 0.0),
                    // height
                    toReal(values[40], 2.5),
                    // x scale
                    toReal(values[41], 1.0),
                    // generation flags
                    toInt(values[71], 0),
                    // h just
                    toInt(values[72], 0),
                    // v just
                    toInt(values[74], 0),
                    // text
                    values[1],
                    // style
                    values[7],
                    // angle
                    (toReal(values[50], 0.0)*2*M_PI)/360.0);

                creationInterface->addText(d);
            }
            break;

        case DL_ENTITY_DIMENSION: {
                int type = (toInt(values[70], 0)&0x07);

                // generic dimension data:
                DL_DimensionData d(
                    // def point
                    toReal(values[10], 0.0),
                    toReal(values[20], 0.0),
                    toReal(values[30], 0.0),
                    // text middle point
                    toReal(values[11], 0.0),
                    toReal(values[21], 0.0),
                    toReal(values[31], 0.0),
                    // type
                    toInt(values[70], 0),
                    // attachment point
                    toInt(values[71], 5),
                    // line sp. style
                    toInt(values[72], 1),
                    // line sp. factor
                    toReal(values[41], 1.0),
                    // text
                    values[1],
                    // style
                    values[3],
                    // angle
                    toReal(values[53], 0.0));

                switch (type) {
                case 0: {
                        // horizontal / vertical / rotated dimension:
                        DL_DimLinearData dl(
                            // definition point 1
                            toReal(values[13], 0.0),
                            toReal(values[23], 0.0),
                            toReal(values[33], 0.0),
                            // definition point 2
                            toReal(values[14], 0.0),
                            toReal(values[24], 0.0),
                            toReal(values[34], 0.0),
                            // angle
                            toReal(values[50], 0.0),
                            // oblique
                            toReal(values[52], 0.0));
                        creationInterface->addDimLinear(d, dl);
                    }
                    break;

                case 1: {
                        // aligned dimension:
                        DL_DimAlignedData da(
                            // extension point 1
                            toReal(values[13], 0.0),
                            toReal(values[23], 0.0),
                            toReal(values[33], 0.0),
                            // extension point 2
                            toReal(values[14], 0.0),
                            toReal(values[24], 0.0),
                            toReal(values[34], 0.0));
                        creationInterface->addDimAlign(d, da);
                    }
                    break;

                case 4: {
                        // radial dimension:
                        DL_DimRadialData dr(
                            // definition point
                            toReal(values[15], 0.0),
                            toReal(values[25], 0.0),
                            toReal(values[35], 0.0),
                            // leader length:
                            toReal(values[40], 0.0));
                        creationInterface->addDimRadial(d, dr);
                    }
                    break;

                case 3: {
                        // diametric dimension:
                        DL_DimDiametricData dr(
                            // definition point
                            toReal(values[15], 0.0),
                            toReal(values[25], 0.0),
                            toReal(values[35], 0.0),
                            // leader length:
                            toReal(values[40], 0.0));
                        creationInterface->addDimDiametric(d, dr);
                    }
                    break;

                case 2: {
                        // angular dimension:
                        DL_DimAngularData da(
                            // definition point 1
                            toReal(values[13], 0.0),
                            toReal(values[23], 0.0),
                            toReal(values[33], 0.0),
                            // definition point 2
                            toReal(values[14], 0.0),
                            toReal(values[24], 0.0),
                            toReal(values[34], 0.0),
                            // definition point 3
                            toReal(values[15], 0.0),
                            toReal(values[25], 0.0),
                            toReal(values[35], 0.0),
                            // definition point 4
                            toReal(values[16], 0.0),
                            toReal(values[26], 0.0),
                            toReal(values[36], 0.0));
                        creationInterface->addDimAngular(d, da);
                    }
                    break;

                case 5: {
                        // angular dimension (3P):
                        DL_DimAngular3PData da(
                            // definition point 1
                            toReal(values[13], 0.0),
                            toReal(values[23], 0.0),
                            toReal(values[33], 0.0),
                            // definition point 2
                            toReal(values[14], 0.0),
                            toReal(values[24], 0.0),
                            toReal(values[34], 0.0),
                            // definition point 3
                            toReal(values[15], 0.0),
                            toReal(values[25], 0.0),
                            toReal(values[35], 0.0));
                        creationInterface->addDimAngular3P(d, da);
                    }
                    break;

                default:
                    break;
                }
            }
            break;

        case DL_ENTITY_LEADER: {
                // leader (arrow)
                DL_LeaderData le(
                    // arrow head flag
                    toInt(values[71], 1),
                    // leader path type
                    toInt(values[72], 0),
                    // Leader creation flag
                    toInt(values[73], 3),
                    // Hookline direction flag
                    toInt(values[74], 1),
                    // Hookline flag
                    toInt(values[75], 0),
                    // Text annotation height
                    toReal(values[40], 1.0),
                    // Text annotation width
                    toReal(values[41], 1.0),
                    // Number of vertices in leader
                    toInt(values[76], 0)
                );
                creationInterface->addLeader(le);

                for (int i=0; i<maxLeaderVertices; i++) {
                    DL_LeaderVertexData d(leaderVertices[i*3],
                                          leaderVertices[i*3+1],
                                          leaderVertices[i*3+2]);

                    creationInterface->addLeaderVertex(d);
                }
            }
            break;

        case DL_ENTITY_HATCH: {
                DL_HatchData hd(toInt(values[91], 1),
                                toInt(values[70], 0),
                                toReal(values[41], 1.0),
                                toReal(values[52], 0.0),
                                values[2]);
                creationInterface->addHatch(hd);

                for (int l=0; l<maxHatchLoops; l++) {
                    DL_HatchLoopData ld(maxHatchEdges[l]);
                    creationInterface->addHatchLoop(ld);
                    for (int b=0; b<maxHatchEdges[l]; b++) {
                        creationInterface->addHatchEdge(hatchEdges[l][b]);
                    }
                }
                creationInterface->endEntity();
                currentEntity = DL_UNKNOWN;
            }
            break;

        case DL_ENTITY_IMAGE: {
                DL_ImageData id(// pass ref insead of name we don't have yet
                    values[340],
                    // ins point:
                    toReal(values[10], 0.0),
                    toReal(values[20], 0.0),
                    toReal(values[30], 0.0),
                    // u vector:
                    toReal(values[11], 1.0),
                    toReal(values[21], 0.0),
                    toReal(values[31], 0.0),
                    // v vector:
                    toReal(values[12], 0.0),
                    toReal(values[22], 1.0),
                    toReal(values[32], 0.0),
                    // image size (pixel):
                    toInt(values[13], 1),
                    toInt(values[23], 1),
                    // brightness, contrast, fade
                    toInt(values[281], 50),
                    toInt(values[282], 50),
                    toInt(values[283], 0));

                creationInterface->addImage(id);
                creationInterface->endEntity();
                currentEntity = DL_UNKNOWN;
            }
            break;

        case DL_ENTITY_IMAGEDEF: {
                DL_ImageDefData id(// handle
                    values[5],
                    values[1]);

                creationInterface->linkImage(id);
                creationInterface->endEntity();
                currentEntity = DL_UNKNOWN;
            }
            break;

        case DL_ENTITY_SEQEND: {
                creationInterface->endSequence();
            }
            break;

        default:
            break;
        }


        // reset all values (they are not persistent and only this
        //  way we can detect default values for unstored settings)
        for (int i=0; i<DL_DXF_MAXGROUPCODE; ++i) {
            values[i][0] = '\0';
        }
        settingValue[0] = '\0';
        settingKey[0] = '\0';


        // Last DXF entity or setting has been handled
        // Now determine what the next entity or setting type is

        // Read DXF settings:
        if (groupValue[0]=='$') {
            currentEntity = DL_SETTING;
            strncpy(settingKey, groupValue, DL_DXF_MAXLINE);
            settingKey[DL_DXF_MAXLINE] = '\0';
        }
        // Read Layers:
        else if (!strcmp(groupValue, "LAYER")) {
            currentEntity = DL_LAYER;

        }
        // Read Blocks:
        else if (!strcmp(groupValue, "BLOCK")) {
            currentEntity = DL_BLOCK;
        } else if (!strcmp(groupValue, "ENDBLK")) {
            currentEntity = DL_ENDBLK;

        }
        // Read entities:
        else if (!strcmp(groupValue, "POINT")) {
            currentEntity = DL_ENTITY_POINT;
        } else if (!strcmp(groupValue, "LINE")) {
            currentEntity = DL_ENTITY_LINE;
        } else if (!strcmp(groupValue, "POLYLINE")) {
            currentEntity = DL_ENTITY_POLYLINE;
        } else if (!strcmp(groupValue, "LWPOLYLINE")) {
            currentEntity = DL_ENTITY_LWPOLYLINE;
        } else if (!strcmp(groupValue, "VERTEX")) {
            currentEntity = DL_ENTITY_VERTEX;
        } else if (!strcmp(groupValue, "ARC")) {
            currentEntity = DL_ENTITY_ARC;
        } else if (!strcmp(groupValue, "ELLIPSE")) {
            currentEntity = DL_ENTITY_ELLIPSE;
        } else if (!strcmp(groupValue, "CIRCLE")) {
            currentEntity = DL_ENTITY_CIRCLE;
        } else if (!strcmp(groupValue, "INSERT")) {
            currentEntity = DL_ENTITY_INSERT;
        } else if (!strcmp(groupValue, "TEXT")) {
            currentEntity = DL_ENTITY_TEXT;
        } else if (!strcmp(groupValue, "MTEXT")) {
            currentEntity = DL_ENTITY_MTEXT;
        } else if (!strcmp(groupValue, "ATTRIB")) {
            currentEntity = DL_ENTITY_ATTRIB;
        } else if (!strcmp(groupValue, "DIMENSION")) {
            currentEntity = DL_ENTITY_DIMENSION;
        } else if (!strcmp(groupValue, "LEADER")) {
            currentEntity = DL_ENTITY_LEADER;
        } else if (!strcmp(groupValue, "HATCH")) {
            currentEntity = DL_ENTITY_HATCH;
        } else if (!strcmp(groupValue, "IMAGE")) {
            currentEntity = DL_ENTITY_IMAGE;
        } else if (!strcmp(groupValue, "IMAGEDEF")) {
            currentEntity = DL_ENTITY_IMAGEDEF;
        } else if (!strcmp(groupValue, "SEQEND")) {
            currentEntity = DL_ENTITY_SEQEND;
        } else {
            currentEntity = DL_UNKNOWN;
        }

        return true;

    } else {
        // Group code does not indicate start of new entity or setting,
        // so this group must be continuation of data for the current
        // one.

        // Read in all codes (no matter if needed or not):
        if (groupCode<DL_DXF_MAXGROUPCODE) {

            // Special handling of text chunks for MTEXT entities:
            if (currentEntity==DL_ENTITY_MTEXT && groupCode==3) {
                creationInterface->addMTextChunk(groupValue);
            }

            // Allocate LWPolyline vertices (group code 90):
            else if (currentEntity==DL_ENTITY_LWPOLYLINE &&
                     groupCode==90) {
                maxVertices = toInt(groupValue);
                if (maxVertices>0) {
                    if (vertices!=NULL) {
                        delete[] vertices;
                    }
                    vertices = new double[4*maxVertices];
                    for (int i=0; i<maxVertices; ++i) {
                        vertices[i*4] = 0.0;
                        vertices[i*4+1] = 0.0;
                        vertices[i*4+2] = 0.0;
                        vertices[i*4+3] = 0.0;
                    }
                }
                vertexIndex=-1;
            }

            // Compute LWPolylines vertices (group codes 10/20/30/42):
            else if (currentEntity==DL_ENTITY_LWPOLYLINE &&
                     (groupCode==10 || groupCode==20 ||
                      groupCode==30 || groupCode==42)) {

                if (vertexIndex<maxVertices-1 && groupCode==10) {
                    vertexIndex++;
                }

                if (groupCode<=30) {
                    if (vertexIndex>=0 && vertexIndex<maxVertices) {
                        vertices[4*vertexIndex + (groupCode/10-1)]
                        = toReal(groupValue);
                    }
                } else if (groupCode==42 && vertexIndex<maxVertices-1) {
                    vertices[4*(vertexIndex+1) + 3] = toReal(groupValue);
                }
            }

            // Allocate Leader vertices (group code 76):
            else if (currentEntity==DL_ENTITY_LEADER &&
                     groupCode==76) {
                maxLeaderVertices = toInt(groupValue);
                if (maxLeaderVertices>0) {
                    if (leaderVertices!=NULL) {
                        delete[] leaderVertices;
                    }
                    leaderVertices = new double[3*maxLeaderVertices];
                    for (int i=0; i<maxLeaderVertices; ++i) {
                        leaderVertices[i*3] = 0.0;
                        leaderVertices[i*3+1] = 0.0;
                        leaderVertices[i*3+2] = 0.0;
                    }
                }
                leaderVertexIndex=-1;
            }

            // Compute Leader vertices (group codes 10/20/30):
            else if (currentEntity==DL_ENTITY_LEADER &&
                     (groupCode==10 || groupCode==20 || groupCode==30)) {

                if (leaderVertexIndex<maxLeaderVertices-1 && groupCode==10) {
                    leaderVertexIndex++;
                }

                if (groupCode<=30) {
                    if (leaderVertexIndex>=0 &&
                            leaderVertexIndex<maxLeaderVertices) {
                        leaderVertices[3*leaderVertexIndex + (groupCode/10-1)]
                        = toReal(groupValue);
                    }
                }
            }

            // Allocate hatch loops (group code 91):
            else if (currentEntity==DL_ENTITY_HATCH &&
                     groupCode==91 && toInt(groupValue)>0) {

                //std::cout << "allocating " << toInt(groupValue) << " loops\n";
                if (hatchLoops!=NULL) {
                    delete[] hatchLoops;
                    hatchLoops = NULL;
                }
                if (maxHatchEdges!=NULL) {
                    delete[] maxHatchEdges;
                    maxHatchEdges = NULL;
                }
                if (hatchEdgeIndex!=NULL) {
                    delete[] hatchEdgeIndex;
                    hatchEdgeIndex = NULL;
                }
                if (hatchEdges!=NULL) {
                    for (int i=0; i<maxHatchLoops; ++i) {
                        delete[] hatchEdges[i];
                    }
                    delete[] hatchEdges;
                    hatchEdges = NULL;
                }
                maxHatchLoops = toInt(groupValue);

				//std::cout << "maxHatchLoops: " << maxHatchLoops << "\n";

                if (maxHatchLoops>0) {
                    hatchLoops = new DL_HatchLoopData[maxHatchLoops];
                    maxHatchEdges = new int[maxHatchLoops];
                    hatchEdgeIndex = new int[maxHatchLoops];
                    hatchEdges = new DL_HatchEdgeData*[maxHatchLoops];
					//std::cout << "new hatchEdges[" << maxHatchLoops << "]\n";
                    for (int i=0; i<maxHatchLoops; ++i) {
                        hatchEdges[i] = NULL;
						//std::cout << "hatchEdges[" << i << "] = NULL\n";
                        maxHatchEdges[i] = 0;
                    }
                    hatchLoopIndex = -1;
					dropEdges = false;
                }
                //std::cout << "done\n";
            }

            // Allocate hatch edges (group code 93)
            else if (currentEntity==DL_ENTITY_HATCH &&
                     groupCode==93 && toInt(groupValue)>0) {

                //std::cout << "  allocating " << toInt(groupValue) << " edges\n";

                if (hatchLoopIndex<maxHatchLoops-1 && hatchLoops!=NULL &&
                        maxHatchEdges!=NULL && hatchEdgeIndex!=NULL &&
                        hatchEdges!=NULL) {

                    hatchLoopIndex++;
                    hatchLoops[hatchLoopIndex]
                    = DL_HatchLoopData(toInt(groupValue));

                    maxHatchEdges[hatchLoopIndex] = toInt(groupValue);
                    hatchEdgeIndex[hatchLoopIndex] = -1;
                   	hatchEdges[hatchLoopIndex]
                   		= new DL_HatchEdgeData[toInt(groupValue)];
					//std::cout << "hatchEdges[" << hatchLoopIndex << "] = new "
					//  << toInt(groupValue) << "\n";
                }
				else {
					dropEdges = true;
				}
                //std::cout << "done\n";
            }

            // Init hatch edge (group code 72)
            else if (currentEntity==DL_ENTITY_HATCH &&
                     hatchEdges!=NULL &&
                     hatchEdgeIndex!=NULL &&
                     maxHatchEdges!=NULL &&
                     hatchLoopIndex>=0 &&
                     hatchLoopIndex<maxHatchLoops &&
                     hatchEdgeIndex[hatchLoopIndex] <
                        maxHatchEdges[hatchLoopIndex] &&
                     (atoi(values[92])&2)==0 &&
                     groupCode==72 &&
					 !dropEdges) {

				//std::cout << "hatchLoopIndex: " << hatchLoopIndex << "\n";
				//std::cout << "maxHatchLoops: " << maxHatchLoops << "\n";

                hatchEdgeIndex[hatchLoopIndex]++;

                //std::cout << "  init edge: type: "
                //<< toInt(groupValue)
                //<< " index: " << hatchEdgeIndex[hatchLoopIndex] << "\n";

                hatchEdges[hatchLoopIndex][hatchEdgeIndex[hatchLoopIndex]]
                .type = toInt(groupValue);
                hatchEdges[hatchLoopIndex][hatchEdgeIndex[hatchLoopIndex]]
                .defined = false;

                //std::cout << "done\n";
            }

            // Handle hatch edges (group codes 10, 20, 11, 21, 40, 50, 51, 73)
            else if (currentEntity==DL_ENTITY_HATCH &&
			         !dropEdges &&
                     hatchEdges!=NULL &&
                     hatchEdgeIndex!=NULL &&
                     hatchLoopIndex>=0 &&
                     hatchLoopIndex<maxHatchLoops &&
                     hatchEdges[hatchLoopIndex]!=NULL &&
                     hatchEdgeIndex[hatchLoopIndex]>=0 &&
                     hatchEdgeIndex[hatchLoopIndex] <
                        maxHatchEdges[hatchLoopIndex] &&
                     ((atoi(values[92])&2)==0) &&        // not a polyline
                     (groupCode==10 || groupCode==20 ||
                      groupCode==11 || groupCode==21 ||
                      groupCode==40 || groupCode==50 ||
                      groupCode==51 || groupCode==73)) {

                //std::cout << "  found edge data: " << groupCode << "\n";
                //std::cout << "     value: " << toReal(groupValue) << "\n";

                // can crash:
                //std::cout << "     defined: "
                //   << (int)hatchEdges[hatchLoopIndex]
                //   [hatchEdgeIndex[hatchLoopIndex]].defined << "\n";

                //std::cout << "92 flag: '" << values[92] << "'\n";
                //std::cout << "92 flag (int): '" << atoi(values[92]) << "'\n";

                if (hatchEdges[hatchLoopIndex]
                        [hatchEdgeIndex[hatchLoopIndex]].defined==false) {
                    if (hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].type==1) {
                        switch (groupCode) {
                        case 10:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].x1
                            = toReal(groupValue);
                            break;
                        case 20:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].y1
                            = toReal(groupValue);
                            break;
                        case 11:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].x2
                            = toReal(groupValue);
                            break;
                        case 21:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].y2
                            = toReal(groupValue);
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].defined = true;
                            break;
                        default:
                            break;
                        }
                    }

                    if (hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].type==2) {
                        switch (groupCode) {
                        case 10:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].cx
                            = toReal(groupValue);
                            break;
                        case 20:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].cy
                            = toReal(groupValue);
                            break;
                        case 40:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].radius
                            = toReal(groupValue);
                            break;
                        case 50:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].angle1
                            = toReal(groupValue)/360.0*2*M_PI;
                            break;
                        case 51:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].angle2
                            = toReal(groupValue)/360.0*2*M_PI;
                            break;
                        case 73:
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].ccw
                            = (bool)toInt(groupValue);
                            hatchEdges[hatchLoopIndex]
                            [hatchEdgeIndex[hatchLoopIndex]].defined = true;
                            break;
                        default:
                            break;
                        }
                    }
                }
            }

            // Normal group / value pair:
            else {
                strncpy(values[groupCode], groupValue, DL_DXF_MAXLINE);
                values[groupCode][DL_DXF_MAXLINE] = '\0';
            }
        }

        return false;
    }
}




/**
 * @brief Opens the given file for writing and returns a pointer
 * to the dxf writer. This pointer needs to be passed on to other
 * writing functions.
 *
 * @param file Full path of the file to open.
 *
 * @return Pointer to an ascii dxf writer object.
 */
DL_WriterA* DL_Dxf::out(const char* file, DL_Codes::version version) {
    char* f = new char[strlen(file)+1];
    strcpy(f, file);
    this->version = version;

    DL_WriterA* dw = new DL_WriterA(f, version);
    return dw;
}



/**
 * @brief Writes a DXF header to the file currently opened 
 * by the given DXF writer object.
 */
void DL_Dxf::writeHeader(DL_WriterA& dw) {
    dw.comment("dxflib " DL_VERSION);
    dw.sectionHeader();

    dw.dxfString(9, "$ACADVER");
    switch (version) {
    case DL_Codes::AC1009:
        dw.dxfString(1, "AC1009");
        break;
    case DL_Codes::AC1012:
        dw.dxfString(1, "AC1012");
        break;
    case DL_Codes::AC1014:
        dw.dxfString(1, "AC1014");
        break;
    case DL_Codes::AC1015:
        dw.dxfString(1, "AC1015");
        break;
    }

    // Newer version require that (otherwise a*cad crashes..)
    if (version==VER_2000) {
        dw.dxfString(9, "$HANDSEED");
        dw.dxfHex(5, 0xFFFF);
    }

    dw.sectionEnd();
}




/**
 * Writes a point entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writePoint(DL_WriterA& dw,
                        const DL_PointData& data,
                        const DL_Attributes& attrib) {
    dw.entity("POINT");
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbPoint");
    }
    dw.entityAttributes(attrib);
    dw.coord(POINT_COORD_CODE, data.x, data.y);
}



/**
 * Writes a line entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeLine(DL_WriterA& dw,
                       const DL_LineData& data,
                       const DL_Attributes& attrib) {
    dw.entity("LINE");
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbLine");
    }
    dw.entityAttributes(attrib);
    dw.coord(LINE_START_CODE, data.x1, data.y1);
    dw.coord(LINE_END_CODE, data.x2, data.y2);
}



/**
 * Writes a polyline entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 * @see writeVertex
 */
void DL_Dxf::writePolyline(DL_WriterA& dw,
                           const DL_PolylineData& data,
                           const DL_Attributes& attrib) {
    if (version==VER_2000) {
        dw.entity("LWPOLYLINE");
        dw.entityAttributes(attrib);
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbPolyline");
        dw.dxfInt(90, (int)data.number);
        dw.dxfInt(70, data.flags);
    } else {
        dw.entity("POLYLINE");
        dw.entityAttributes(attrib);
        dw.coord(VERTEX_COORD_CODE, 0.0, 0.0);
    }
}



/**
 * Writes a single vertex of a polyline to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeVertex(DL_WriterA& dw,
                         const DL_VertexData& data) {

	
    if (version==VER_2000) {
        if (fabs(data.bulge)>1.0e-10) {
            dw.dxfReal(42, data.bulge);
        }
        dw.dxfReal(10, data.x);
        dw.dxfReal(20, data.y);
    } else {
        dw.entity("VERTEX");
        //dw.entityAttributes(attrib);
        dw.coord(VERTEX_COORD_CODE, data.x, data.y);
        if (fabs(data.bulge)>1.0e-10) {
            dw.dxfReal(42, data.bulge);
        }
    }
}



/**
 * Writes a circle entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeCircle(DL_WriterA& dw,
                         const DL_CircleData& data,
                         const DL_Attributes& attrib) {
    dw.entity("CIRCLE");
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbCircle");
    }
    dw.entityAttributes(attrib);
    dw.coord(10, data.cx, data.cy);
    dw.dxfReal(40, data.radius);
}



/**
 * Writes an arc entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeArc(DL_WriterA& dw,
                      const DL_ArcData& data,
                      const DL_Attributes& attrib) {
    dw.entity("ARC");
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
    }
    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbCircle");
    }
    dw.coord(10, data.cx, data.cy);
    dw.dxfReal(40, data.radius);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbArc");
    }
    dw.dxfReal(50, data.angle1);
    dw.dxfReal(51, data.angle2);
}



/**
 * Writes an ellipse entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeEllipse(DL_WriterA& dw,
                          const DL_EllipseData& data,
                          const DL_Attributes& attrib) {

    dw.entity("ELLIPSE");
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbEllipse");
    }
    dw.entityAttributes(attrib);
    dw.coord(10, data.cx, data.cy);
    dw.coord(11, data.mx, data.my);
    dw.dxfReal(40, data.ratio);
    dw.dxfReal(41, data.angle1);
    dw.dxfReal(42, data.angle2);
}



/**
 * Writes an insert to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeInsert(DL_WriterA& dw,
                         const DL_InsertData& data,
                         const DL_Attributes& attrib) {

    if (data.name.empty()) {
        std::cerr << "DL_Dxf::writeInsert: "
        << "Block name must not be empty\n";
        return;
    }

    dw.entity("INSERT");
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbBlockReference");
    }
    dw.entityAttributes(attrib);
    dw.dxfString(2, data.name);
    dw.dxfReal(10, data.ipx);
    dw.dxfReal(20, data.ipy);
    dw.dxfReal(30, 0.0);
    if (data.sx!=1.0 || data.sy!=1.0) {
        dw.dxfReal(41, data.sx);
        dw.dxfReal(42, data.sy);
        dw.dxfReal(43, 1.0);
    }
    if (data.angle!=0.0) {
        dw.dxfReal(50, data.angle);
    }
    if (data.cols!=1 || data.rows!=1) {
        dw.dxfInt(70, data.cols);
        dw.dxfInt(71, data.rows);
    }
    if (data.colSp!=0.0 || data.rowSp!=0.0) {
        dw.dxfReal(44, data.colSp);
        dw.dxfReal(45, data.rowSp);
    }

}



/**
 * Writes a text entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeMText(DL_WriterA& dw,
                        const DL_MTextData& data,
                        const DL_Attributes& attrib) {

    dw.entity("MTEXT");
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbMText");
    }
    dw.entityAttributes(attrib);
    dw.dxfReal(10, data.ipx);
    dw.dxfReal(20, data.ipy);
    dw.dxfReal(30, 0.0);
    dw.dxfReal(40, data.height);
    dw.dxfReal(41, data.width);

    dw.dxfInt(71, data.attachmentPoint);
    dw.dxfInt(72, data.drawingDirection);

    // Creare text chunks of 250 characters each:
    int length = data.text.length();
    char chunk[251];
    int i;
    for (i=250; i<length; i+=250) {
        strncpy(chunk, &data.text.c_str()[i-250], 250);
        chunk[250]='\0';
        dw.dxfString(3, chunk);
    }
    strncpy(chunk, &data.text.c_str()[i-250], 250);
    chunk[250]='\0';
    dw.dxfString(1, chunk);

    dw.dxfString(7, data.style);

    dw.dxfReal(50, data.angle);

    dw.dxfInt(73, data.lineSpacingStyle);
    dw.dxfReal(44, data.lineSpacingFactor);
}



/**
 * Writes an aligned dimension entity to the file.
 *
 * @param dw DXF writer
 * @param data Generic dimension data for from the file
 * @param data Specific aligned dimension data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeDimAligned(DL_WriterA& dw,
                             const DL_DimensionData& data,
                             const DL_DimAlignedData& edata,
                             const DL_Attributes& attrib) {

    dw.entity("DIMENSION");

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
    }
    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbDimension");
    }

    dw.dxfReal(10, data.dpx);
    dw.dxfReal(20, data.dpy);
    dw.dxfReal(30, 0.0);

    dw.dxfReal(11, data.mpx);
    dw.dxfReal(21, data.mpy);
    dw.dxfReal(31, 0.0);

    dw.dxfInt(70, 1);
    dw.dxfInt(71, data.attachmentPoint);
    dw.dxfInt(72, data.lineSpacingStyle); // opt
    dw.dxfReal(41, data.lineSpacingFactor); // opt

    dw.dxfReal(42, data.angle);

    dw.dxfString(1, data.text);   // opt
    dw.dxfString(3, data.style);

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbAlignedDimension");
    }

    dw.dxfReal(13, edata.epx1);
    dw.dxfReal(23, edata.epy1);
    dw.dxfReal(33, 0.0);

    dw.dxfReal(14, edata.epx2);
    dw.dxfReal(24, edata.epy2);
    dw.dxfReal(34, 0.0);
}



/**
 * Writes a linear dimension entity to the file.
 *
 * @param dw DXF writer
 * @param data Generic dimension data for from the file
 * @param data Specific linear dimension data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeDimLinear(DL_WriterA& dw,
                            const DL_DimensionData& data,
                            const DL_DimLinearData& edata,
                            const DL_Attributes& attrib) {

    dw.entity("DIMENSION");

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
    }
    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbDimension");
    }

    dw.dxfReal(10, data.dpx);
    dw.dxfReal(20, data.dpy);
    dw.dxfReal(30, 0.0);

    dw.dxfReal(11, data.mpx);
    dw.dxfReal(21, data.mpy);
    dw.dxfReal(31, 0.0);

    dw.dxfInt(70, 0);
    dw.dxfInt(71, data.attachmentPoint);
    dw.dxfInt(72, data.lineSpacingStyle); // opt
    dw.dxfReal(41, data.lineSpacingFactor); // opt

    dw.dxfReal(42, data.angle);

    dw.dxfString(1, data.text);   // opt
    dw.dxfString(3, data.style);

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbAlignedDimension");
    }

    dw.dxfReal(13, edata.dpx1);
    dw.dxfReal(23, edata.dpy1);
    dw.dxfReal(33, 0.0);

    dw.dxfReal(14, edata.dpx2);
    dw.dxfReal(24, edata.dpy2);
    dw.dxfReal(34, 0.0);

    dw.dxfReal(50, edata.angle/(2.0*M_PI)*360.0);

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbRotatedDimension");
    }
}



/**
 * Writes a radial dimension entity to the file.
 *
 * @param dw DXF writer
 * @param data Generic dimension data for from the file
 * @param data Specific radial dimension data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeDimRadial(DL_WriterA& dw,
                            const DL_DimensionData& data,
                            const DL_DimRadialData& edata,
                            const DL_Attributes& attrib) {

    dw.entity("DIMENSION");

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
    }
    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbDimension");
    }

    dw.dxfReal(10, data.dpx);
    dw.dxfReal(20, data.dpy);
    dw.dxfReal(30, 0.0);

    dw.dxfReal(11, data.mpx);
    dw.dxfReal(21, data.mpy);
    dw.dxfReal(31, 0.0);

    dw.dxfInt(70, 4);
    dw.dxfInt(71, data.attachmentPoint);
    dw.dxfInt(72, data.lineSpacingStyle); // opt
    dw.dxfReal(41, data.lineSpacingFactor); // opt

    dw.dxfReal(42, data.angle);

    dw.dxfString(1, data.text);   // opt
    dw.dxfString(3, data.style);

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbRadialDimension");
    }

    dw.dxfReal(15, edata.dpx);
    dw.dxfReal(25, edata.dpy);
    dw.dxfReal(35, 0.0);

    dw.dxfReal(40, edata.leader);
}



/**
 * Writes a diametric dimension entity to the file.
 *
 * @param dw DXF writer
 * @param data Generic dimension data for from the file
 * @param data Specific diametric dimension data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeDimDiametric(DL_WriterA& dw,
                               const DL_DimensionData& data,
                               const DL_DimDiametricData& edata,
                               const DL_Attributes& attrib) {

    dw.entity("DIMENSION");

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
    }
    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbDimension");
    }

    dw.dxfReal(10, data.dpx);
    dw.dxfReal(20, data.dpy);
    dw.dxfReal(30, 0.0);

    dw.dxfReal(11, data.mpx);
    dw.dxfReal(21, data.mpy);
    dw.dxfReal(31, 0.0);

    dw.dxfInt(70, 3);
    dw.dxfInt(71, data.attachmentPoint);
    dw.dxfInt(72, data.lineSpacingStyle); // opt
    dw.dxfReal(41, data.lineSpacingFactor); // opt

    dw.dxfReal(42, data.angle);

    dw.dxfString(1, data.text);   // opt
    dw.dxfString(3, data.style);

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbDiametricDimension");
    }

    dw.dxfReal(15, edata.dpx);
    dw.dxfReal(25, edata.dpy);
    dw.dxfReal(35, 0.0);

    dw.dxfReal(40, edata.leader);
}



/**
 * Writes an angular dimension entity to the file.
 *
 * @param dw DXF writer
 * @param data Generic dimension data for from the file
 * @param data Specific angular dimension data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeDimAngular(DL_WriterA& dw,
                             const DL_DimensionData& data,
                             const DL_DimAngularData& edata,
                             const DL_Attributes& attrib) {

    dw.entity("DIMENSION");

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
    }
    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbDimension");
    }

    dw.dxfReal(10, data.dpx);
    dw.dxfReal(20, data.dpy);
    dw.dxfReal(30, 0.0);

    dw.dxfReal(11, data.mpx);
    dw.dxfReal(21, data.mpy);
    dw.dxfReal(31, 0.0);

    dw.dxfInt(70, 2);
    dw.dxfInt(71, data.attachmentPoint);
    dw.dxfInt(72, data.lineSpacingStyle); // opt
    dw.dxfReal(41, data.lineSpacingFactor); // opt

    dw.dxfReal(42, data.angle);

    dw.dxfString(1, data.text);   // opt
    dw.dxfString(3, data.style);

    if (version==VER_2000) {
        dw.dxfString(100, "AcDb2LineAngularDimension");
    }

    dw.dxfReal(13, edata.dpx1);
    dw.dxfReal(23, edata.dpy1);
    dw.dxfReal(33, 0.0);

    dw.dxfReal(14, edata.dpx2);
    dw.dxfReal(24, edata.dpy2);
    dw.dxfReal(34, 0.0);

    dw.dxfReal(15, edata.dpx3);
    dw.dxfReal(25, edata.dpy3);
    dw.dxfReal(35, 0.0);

    dw.dxfReal(16, edata.dpx4);
    dw.dxfReal(26, edata.dpy4);
    dw.dxfReal(36, 0.0);
}



/**
 * Writes an angular dimension entity (3 points version) to the file.
 *
 * @param dw DXF writer
 * @param data Generic dimension data for from the file
 * @param data Specific angular dimension data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeDimAngular3P(DL_WriterA& dw,
                               const DL_DimensionData& data,
                               const DL_DimAngular3PData& edata,
                               const DL_Attributes& attrib) {

    dw.entity("DIMENSION");

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
    }
    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbDimension");
    }

    dw.dxfReal(10, data.dpx);
    dw.dxfReal(20, data.dpy);
    dw.dxfReal(30, 0.0);

    dw.dxfReal(11, data.mpx);
    dw.dxfReal(21, data.mpy);
    dw.dxfReal(31, 0.0);

    dw.dxfInt(70, 5);
    dw.dxfInt(71, data.attachmentPoint);
    dw.dxfInt(72, data.lineSpacingStyle); // opt
    dw.dxfReal(41, data.lineSpacingFactor); // opt

    dw.dxfReal(42, data.angle);

    dw.dxfString(1, data.text);   // opt
    dw.dxfString(3, data.style);

    if (version==VER_2000) {
        dw.dxfString(100, "AcDb3PointAngularDimension");
    }

    dw.dxfReal(13, edata.dpx1);
    dw.dxfReal(23, edata.dpy1);
    dw.dxfReal(33, 0.0);

    dw.dxfReal(14, edata.dpx2);
    dw.dxfReal(24, edata.dpy2);
    dw.dxfReal(34, 0.0);

    dw.dxfReal(15, edata.dpx3);
    dw.dxfReal(25, edata.dpy3);
    dw.dxfReal(35, 0.0);
}



/**
 * Writes a leader entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 * @see writeVertex
 */
void DL_Dxf::writeLeader(DL_WriterA& dw,
                         const DL_LeaderData& data,
                         const DL_Attributes& attrib) {
    dw.entity("LEADER");
    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbLeader");
    }
    dw.dxfString(3, "Standard");
    dw.dxfInt(71, data.arrowHeadFlag);
    dw.dxfInt(72, data.leaderPathType);
    dw.dxfInt(73, data.leaderCreationFlag);
    dw.dxfInt(74, data.hooklineDirectionFlag);
    dw.dxfInt(75, data.hooklineFlag);
    dw.dxfReal(40, data.textAnnotationHeight);
    dw.dxfReal(41, data.textAnnotationWidth);
    dw.dxfInt(76, data.number);
}



/**
 * Writes a single vertex of a leader to the file.
 *
 * @param dw DXF writer
 * @param data Entity data
 */
void DL_Dxf::writeLeaderVertex(DL_WriterA& dw,
                               const DL_LeaderVertexData& data) {
    dw.dxfReal(10, data.x);
    dw.dxfReal(20, data.y);
}



/**
 * Writes the beginning of a hatch entity to the file.
 * This must be followed by one or more writeHatchLoop()
 * calls and a writeHatch2() call.
 *
 * @param dw DXF writer
 * @param data Entity data.
 * @param attrib Attributes
 */
void DL_Dxf::writeHatch1(DL_WriterA& dw,
                         const DL_HatchData& data,
                         const DL_Attributes& attrib) {

    dw.entity("HATCH");
    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbHatch");
    }
    dw.dxfReal(10, 0.0);             // elevation
    dw.dxfReal(20, 0.0);
    dw.dxfReal(30, 0.0);
    dw.dxfReal(210, 0.0);             // extrusion dir.
    dw.dxfReal(220, 0.0);
    dw.dxfReal(230, 1.0);
    if (data.solid==false) {
        dw.dxfString(2, data.pattern);
    } else {
        dw.dxfString(2, "SOLID");
    }
    dw.dxfInt(70, (int)data.solid);
    dw.dxfInt(71, 0);                // associative
    dw.dxfInt(91, data.numLoops);
}



/**
 * Writes the end of a hatch entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data.
 * @param attrib Attributes
 */
void DL_Dxf::writeHatch2(DL_WriterA& dw,
                         const DL_HatchData& data,
                         const DL_Attributes& /*attrib*/) {

    dw.dxfInt(75, 0);                // odd parity
    dw.dxfInt(76, 1);                // pattern type
    if (data.solid==false) {
        dw.dxfReal(52, data.angle);
        dw.dxfReal(41, data.scale);
        dw.dxfInt(77, 0);            // not double
        //dw.dxfInt(78, 0);
        dw.dxfInt(78, 1);
        dw.dxfReal(53, 45.0);
        dw.dxfReal(43, 0.0);
        dw.dxfReal(44, 0.0);
        dw.dxfReal(45, -0.0883883476483184);
        dw.dxfReal(46, 0.0883883476483185);
        dw.dxfInt(79, 0);
    }
    dw.dxfInt(98, 0);
}



/**
 * Writes the beginning of a hatch loop to the file. This
 * must happen after writing the beginning of a hatch entity.
 *
 * @param dw DXF writer
 * @param data Entity data.
 * @param attrib Attributes
 */
void DL_Dxf::writeHatchLoop1(DL_WriterA& dw,
                             const DL_HatchLoopData& data) {

    dw.dxfInt(92, 1);
    dw.dxfInt(93, data.numEdges);
    //dw.dxfInt(97, 0);
}



/**
 * Writes the end of a hatch loop to the file.
 *
 * @param dw DXF writer
 * @param data Entity data.
 * @param attrib Attributes
 */
void DL_Dxf::writeHatchLoop2(DL_WriterA& dw,
                             const DL_HatchLoopData& /*data*/) {

    dw.dxfInt(97, 0);
}


/**
 * Writes the beginning of a hatch entity to the file.
 *
 * @param dw DXF writer
 * @param data Entity data.
 * @param attrib Attributes
 */
void DL_Dxf::writeHatchEdge(DL_WriterA& dw,
                            const DL_HatchEdgeData& data) {

    dw.dxfInt(72, data.type);

    switch (data.type) {
    case 1:
        dw.dxfReal(10, data.x1);
        dw.dxfReal(20, data.y1);
        dw.dxfReal(11, data.x2);
        dw.dxfReal(21, data.y2);
        break;
    case 2:
        dw.dxfReal(10, data.cx);
        dw.dxfReal(20, data.cy);
        dw.dxfReal(40, data.radius);
        dw.dxfReal(50, data.angle1/(2*M_PI)*360.0);
        dw.dxfReal(51, data.angle2/(2*M_PI)*360.0);
        dw.dxfInt(73, (int)(data.ccw));
        break;
    default:
        break;
    }
}



/**
 * Writes an image entity.
 *
 * @return IMAGEDEF handle. Needed for the IMAGEDEF counterpart.
 */
int DL_Dxf::writeImage(DL_WriterA& dw,
                       const DL_ImageData& data,
                       const DL_Attributes& attrib) {

    /*if (data.file.empty()) {
        std::cerr << "DL_Dxf::writeImage: "
        << "Image file must not be empty\n";
        return;
}*/

    dw.entity("IMAGE");

    dw.entityAttributes(attrib);
    if (version==VER_2000) {
        dw.dxfString(100, "AcDbEntity");
        dw.dxfString(100, "AcDbRasterImage");
        dw.dxfInt(90, 0);
    }
    // insertion point
    dw.dxfReal(10, data.ipx);
    dw.dxfReal(20, data.ipy);
    dw.dxfReal(30, 0.0);

    // vector along bottom side (1 pixel long)
    dw.dxfReal(11, data.ux);
    dw.dxfReal(21, data.uy);
    dw.dxfReal(31, 0.0);

    // vector along left side (1 pixel long)
    dw.dxfReal(12, data.vx);
    dw.dxfReal(22, data.vy);
    dw.dxfReal(32, 0.0);

    // image size in pixel
    dw.dxfReal(13, data.width);
    dw.dxfReal(23, data.height);

    // handle of IMAGEDEF object
    int handle = dw.incHandle();
    dw.dxfHex(340, handle);

    // flags
    dw.dxfInt(70, 15);

    // clipping:
    dw.dxfInt(280, 0);

    // brightness, contrast, fade
    dw.dxfInt(281, data.brightness);
    dw.dxfInt(282, data.contrast);
    dw.dxfInt(283, data.fade);

    return handle;
}



/**
 * Writes an image definiition entity.
 */
void DL_Dxf::writeImageDef(DL_WriterA& dw,
                           int handle,
                           const DL_ImageData& data) {

    /*if (data.file.empty()) {
        std::cerr << "DL_Dxf::writeImage: "
        << "Image file must not be empty\n";
        return;
}*/

    dw.dxfString(0, "IMAGEDEF");
    dw.dxfHex(5, handle);

    if (version==VER_2000) {
        dw.dxfString(100, "AcDbRasterImageDef");
        dw.dxfInt(90, 0);
    }
    // file name:
    dw.dxfString(1, data.ref);

    // image size in pixel
    dw.dxfReal(10, data.width);
    dw.dxfReal(20, data.height);

    dw.dxfReal(11, 1.0);
    dw.dxfReal(21, 1.0);

    // loaded:
    dw.dxfInt(280, 1);
    // units:
    dw.dxfInt(281, 0);
}


/**
 * Writes a layer to the file. Layers are stored in the 
 * tables section of a DXF file.
 *
 * @param dw DXF writer
 * @param data Entity data from the file
 * @param attrib Attributes
 */
void DL_Dxf::writeLayer(DL_WriterA& dw,
                        const DL_LayerData& data,
                        const DL_Attributes& attrib) {

    if (data.name.empty()) {
        std::cerr << "DL_Dxf::writeLayer: "
        << "Layer name must not be empty\n";
        return;
    }

    int color = attrib.getColor();
    if (color==0) {
        std::cerr << "Layer color cannot be 0. Corrected to 7.\n";
        color = 7;
    }

    if (data.name == "0") {
        dw.tableLayerEntry(0x10);
    } else {
        dw.tableLayerEntry();
    }

    dw.dxfString(2, data.name);
    dw.dxfInt(70, data.flags);
    dw.dxfInt(62, color);

    dw.dxfString(6, (attrib.getLineType().length()==0 ?
                     string("CONTINUOUS") : attrib.getLineType()));

    if (version>=VER_2000) {
		// layer defpoints cannot be plotted
		std::string lstr = data.name;
		std::transform(lstr.begin(), lstr.end(), lstr.begin(), tolower);
		if (lstr=="defpoints") {
	    	dw.dxfInt(290, 0);
		}
	}
    if (version>=VER_2000 && attrib.getWidth()!=-1) {
        dw.dxfInt(370, attrib.getWidth());
    }
    if (version>=VER_2000) {
	    dw.dxfHex(390, 0xF);
	}
}



/**
 * Writes a line type to the file. Line types are stored in the 
 * tables section of a DXF file.
 */
void DL_Dxf::writeLineType(DL_WriterA& dw,
                           const DL_LineTypeData& data) {
    //const char* description,
    //int elements,
    //double patternLength) {

    if (data.name.empty()) {
        std::cerr << "DL_Dxf::writeLineType: "
        << "Line type name must not be empty\n";
        return;
    }

    if (!strcasecmp(data.name.c_str(), "BYBLOCK")) {
        dw.tableLineTypeEntry(0x14);
    } else if (!strcasecmp(data.name.c_str(), "BYLAYER")) {
        dw.tableLineTypeEntry(0x15);
    } else if (!strcasecmp(data.name.c_str(), "CONTINUOUS")) {
        dw.tableLineTypeEntry(0x16);
    } else {
        dw.tableLineTypeEntry();
    }

    dw.dxfString(2, data.name);
    dw.dxfInt(70, data.flags);

    if (!strcasecmp(data.name.c_str(), "BYBLOCK")) {
        dw.dxfString(3, "");
        dw.dxfInt(72, 65);
        dw.dxfInt(73, 0);
        dw.dxfReal(40, 0.0);
    } else if (!strcasecmp(data.name.c_str(), "BYLAYER")) {
        dw.dxfString(3, "");
        dw.dxfInt(72, 65);
        dw.dxfInt(73, 0);
        dw.dxfReal(40, 0.0);
    } else if (!strcasecmp(data.name.c_str(), "CONTINUOUS")) {
        dw.dxfString(3, "Solid line");
        dw.dxfInt(72, 65);
        dw.dxfInt(73, 0);
        dw.dxfReal(40, 0.0);
    } else if (!strcasecmp(data.name.c_str(), "ACAD_ISO02W100")) {
        dw.dxfString(3, "ISO Dashed __ __ __ __ __ __ __ __ __ __ _");
        dw.dxfInt(72, 65);
        dw.dxfInt(73, 2);
        dw.dxfReal(40, 15.0);
        dw.dxfReal(49, 12.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, -3.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
    } else if (!strcasecmp(data.name.c_str(), "ACAD_ISO03W100")) {
        dw.dxfString(3, "ISO Dashed with Distance __    __    __    _");
        dw.dxfInt(72, 65);
        dw.dxfInt(73, 2);
        dw.dxfReal(40, 30.0);
        dw.dxfReal(49, 12.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, -18.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
    } else if (!strcasecmp(data.name.c_str(), "ACAD_ISO04W100")) {
        dw.dxfString(3, "ISO Long Dashed Dotted ____ . ____ . __");
        dw.dxfInt(72, 65);
        dw.dxfInt(73, 4);
        dw.dxfReal(40, 30.0);
        dw.dxfReal(49, 24.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, -3.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, 0.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, -3.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
    } else if (!strcasecmp(data.name.c_str(), "ACAD_ISO05W100")) {
        dw.dxfString(3, "ISO Long Dashed Double Dotted ____ .. __");
        dw.dxfInt(72, 65);
        dw.dxfInt(73, 6);
        dw.dxfReal(40, 33.0);
        dw.dxfReal(49, 24.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, -3.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, 0.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, -3.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, 0.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
        dw.dxfReal(49, -3.0);
        if (version>=VER_R13)
            dw.dxfInt(74, 0);
    }
}



/**
 * Writes the APPID section to the DXF file.
 *
 * @param name Application name
 */
void DL_Dxf::writeAppid(DL_WriterA& dw, const string& name) {
    if (name.empty()) {
        std::cerr << "DL_Dxf::writeAppid: "
        << "Application  name must not be empty\n";
        return;
    }

    if (!strcasecmp(name.c_str(), "ACAD")) {
        dw.tableAppidEntry(0x12);
    } else {
        dw.tableAppidEntry();
    }
    dw.dxfString(2, name);
    dw.dxfInt(70, 0);
}



/**
 * Writes a block's definition (no entities) to the DXF file.
 */
void DL_Dxf::writeBlock(DL_WriterA& dw, const DL_BlockData& data) {
    if (data.name.empty()) {
        std::cerr << "DL_Dxf::writeBlock: "
        << "Block name must not be empty\n";
        return;
    }

    //bool paperSpace = !strcasecmp(name, "*paper_space");
    //!strcasecmp(name, "*paper_space0");

    if (!strcasecmp(data.name.c_str(), "*paper_space")) {
        dw.sectionBlockEntry(0x1C);
    } else if (!strcasecmp(data.name.c_str(), "*model_space")) {
        dw.sectionBlockEntry(0x20);
    } else if (!strcasecmp(data.name.c_str(), "*paper_space0")) {
        dw.sectionBlockEntry(0x24);
    } else {
        dw.sectionBlockEntry();
    }
    dw.dxfString(2, data.name);
    dw.dxfInt(70, 0);
    dw.coord(10, data.bpx, data.bpy);
    dw.dxfString(3, data.name);
    dw.dxfString(1, "");
}



/**
 * Writes a block end.
 *
 * @param name Block name
 */
void DL_Dxf::writeEndBlock(DL_WriterA& dw, const string& name) {
    if (!strcasecmp(name.c_str(), "*paper_space")) {
        dw.sectionBlockEntryEnd(0x1D);
    } else if (!strcasecmp(name.c_str(), "*model_space")) {
        dw.sectionBlockEntryEnd(0x21);
    } else if (!strcasecmp(name.c_str(), "*paper_space0")) {
        dw.sectionBlockEntryEnd(0x25);
    } else {
        dw.sectionBlockEntryEnd();
    }
}



/**
 * Writes a viewport section. This section is needed in VER_R13.
 * Note that this method currently only writes a faked VPORT section
 * to make the file readable by Aut*cad.
 */
void DL_Dxf::writeVPort(DL_WriterA& dw) {
    dw.dxfString(0, "TABLE");
    dw.dxfString(2, "VPORT");
    dw.dxfHex(5, 0x8);
    //dw.dxfHex(330, 0);
    dw.dxfString(100, "AcDbSymbolTable");
    dw.dxfInt(70, 1);
    dw.dxfString(0, "VPORT");
    //dw.dxfHex(5, 0x2F);
    dw.handle();
    //dw.dxfHex(330, 8);
    dw.dxfString(100, "AcDbSymbolTableRecord");
    dw.dxfString(100, "AcDbViewportTableRecord");
    dw.dxfString(  2, "*Active");
    dw.dxfInt( 70, 0);
    dw.dxfReal( 10, 0.0);
    dw.dxfReal( 20, 0.0);
    dw.dxfReal( 11, 1.0);
    dw.dxfReal( 21, 1.0);
    dw.dxfReal( 12, 286.3055555555555);
    dw.dxfReal( 22, 148.5);
    dw.dxfReal( 13, 0.0);
    dw.dxfReal( 23, 0.0);
    dw.dxfReal( 14, 10.0);
    dw.dxfReal( 24, 10.0);
    dw.dxfReal( 15, 10.0);
    dw.dxfReal( 25, 10.0);
    dw.dxfReal( 16, 0.0);
    dw.dxfReal( 26, 0.0);
    dw.dxfReal( 36, 1.0);
    dw.dxfReal( 17, 0.0);
    dw.dxfReal( 27, 0.0);
    dw.dxfReal( 37, 0.0);
    dw.dxfReal( 40, 297.0);
    dw.dxfReal( 41, 1.92798353909465);
    dw.dxfReal( 42, 50.0);
    dw.dxfReal( 43, 0.0);
    dw.dxfReal( 44, 0.0);
    dw.dxfReal( 50, 0.0);
    dw.dxfReal( 51, 0.0);
    dw.dxfInt( 71, 0);
    dw.dxfInt( 72, 100);
    dw.dxfInt( 73, 1);
    dw.dxfInt( 74, 3);
    dw.dxfInt( 75, 1);
    dw.dxfInt( 76, 1);
    dw.dxfInt( 77, 0);
    dw.dxfInt( 78, 0);
    dw.dxfInt(281, 0);
    dw.dxfInt( 65, 1);
    dw.dxfReal(110, 0.0);
    dw.dxfReal(120, 0.0);
    dw.dxfReal(130, 0.0);
    dw.dxfReal(111, 1.0);
    dw.dxfReal(121, 0.0);
    dw.dxfReal(131, 0.0);
    dw.dxfReal(112, 0.0);
    dw.dxfReal(122, 1.0);
    dw.dxfReal(132, 0.0);
    dw.dxfInt( 79, 0);
    dw.dxfReal(146, 0.0);
    dw.dxfString(  0, "ENDTAB");
}



/**
 * Writes a style section. This section is needed in VER_R13.
 * Note that this method currently only writes a faked STYLE section
 * to make the file readable by Aut*cad.
 */
void DL_Dxf::writeStyle(DL_WriterA& dw) {
    dw.dxfString(  0, "TABLE");
    dw.dxfString(  2, "STYLE");
    dw.dxfHex(5, 3);
    //dw.dxfHex(330, 0);
    dw.dxfString(100, "AcDbSymbolTable");
    dw.dxfInt( 70, 1);
    dw.dxfString(  0, "STYLE");
    dw.dxfHex(5, 0x11);
    //styleHandleStd = dw.handle();
    //dw.dxfHex(330, 3);
    dw.dxfString(100, "AcDbSymbolTableRecord");
    dw.dxfString(100, "AcDbTextStyleTableRecord");
    dw.dxfString(  2, "Standard");
    dw.dxfInt( 70, 0);
    dw.dxfReal( 40, 0.0);
    dw.dxfReal( 41, 1.0);
    dw.dxfReal( 50, 0.0);
    dw.dxfInt( 71, 0);
    dw.dxfReal( 42, 2.5);
    dw.dxfString(  3, "txt");
    dw.dxfString(  4, "");
    dw.dxfString(  0, "ENDTAB");
}



/**
 * Writes a view section. This section is needed in VER_R13.
 * Note that this method currently only writes a faked VIEW section
 * to make the file readable by Aut*cad.
 */
void DL_Dxf::writeView(DL_WriterA& dw) {
    dw.dxfString(  0, "TABLE");
    dw.dxfString(  2, "VIEW");
    dw.dxfHex(5, 6);
    //dw.dxfHex(330, 0);
    dw.dxfString(100, "AcDbSymbolTable");
    dw.dxfInt( 70, 0);
    dw.dxfString(  0, "ENDTAB");
}



/**
 * Writes a ucs section. This section is needed in VER_R13.
 * Note that this method currently only writes a faked UCS section
 * to make the file readable by Aut*cad.
 */
void DL_Dxf::writeUcs(DL_WriterA& dw) {
    dw.dxfString(  0, "TABLE");
    dw.dxfString(  2, "UCS");
    dw.dxfHex(5, 7);
    //dw.dxfHex(330, 0);
    dw.dxfString(100, "AcDbSymbolTable");
    dw.dxfInt( 70, 0);
    dw.dxfString(  0, "ENDTAB");
}



/**
 * Writes a dimstyle section. This section is needed in VER_R13.
 * Note that this method currently only writes a faked DIMSTYLE section
 * to make the file readable by Aut*cad.
 */
void DL_Dxf::writeDimStyle(DL_WriterA& dw) {
    dw.dxfString(  0, "TABLE");
    dw.dxfString(  2, "DIMSTYLE");
    dw.dxfHex(5, 0xA);
    //dw.dxfHex(330, 0);
    dw.dxfString(100, "AcDbSymbolTable");
    dw.dxfInt( 70, 1);
    dw.dxfString(100, "AcDbDimStyleTable");
    dw.dxfInt( 71, 0);
    dw.dxfString(  0, "DIMSTYLE");
    dw.dxfHex(105, 0x27);
    //dw.handle(105);
    //dw.dxfHex(330, 0xA);
    dw.dxfString(100, "AcDbSymbolTableRecord");
    dw.dxfString(100, "AcDbDimStyleTableRecord");
    dw.dxfString(  2, "Standard");
    dw.dxfInt( 70, 0);
	
    //dw.dxfReal(140, 2.5);
    /*
       dw.dxfReal( 41, 2.5);
       dw.dxfReal( 42, 0.625);
       dw.dxfReal( 43, 3.75);
       dw.dxfReal( 44, 1.25);
       dw.dxfInt( 73, 0);
       dw.dxfInt( 74, 0);
       dw.dxfInt( 77, 1);
       dw.dxfInt( 78, 8);
       dw.dxfReal(140, 2.5);
       dw.dxfReal(141, 2.5);
       dw.dxfReal(143, 0.03937007874016);
       dw.dxfReal(147, 0.625);
       dw.dxfInt(171, 3);
       dw.dxfInt(172, 1);
       dw.dxfInt(271, 2);
       dw.dxfInt(272, 2);
       dw.dxfInt(274, 3);
       dw.dxfInt(278, 44);
       dw.dxfInt(283, 0);
       dw.dxfInt(284, 8);
       //dw.dxfHex(340, styleHandleStd);
       dw.dxfHex(340, 0x11);
    */
    dw.dxfString(  0, "ENDTAB");
}



/**
 * Writes a blockrecord section. This section is needed in VER_R13.
 * Note that this method currently only writes a faked BLOCKRECORD section
 * to make the file readable by Aut*cad.
 */
void DL_Dxf::writeBlockRecord(DL_WriterA& dw) {
    dw.dxfString(  0, "TABLE");
    dw.dxfString(  2, "BLOCK_RECORD");
    dw.dxfHex(5, 1);
    //dw.dxfHex(330, 0);
    dw.dxfString(100, "AcDbSymbolTable");
    dw.dxfInt( 70, 1);

    dw.dxfString(  0, "BLOCK_RECORD");
    dw.dxfHex(5, 0x1F);
    //int msh = dw.handle();
    //dw.setModelSpaceHandle(msh);
    //dw.dxfHex(330, 1);
    dw.dxfString(100, "AcDbSymbolTableRecord");
    dw.dxfString(100, "AcDbBlockTableRecord");
    dw.dxfString(  2, "*Model_Space");
    dw.dxfHex(340, 0x22);

    dw.dxfString(  0, "BLOCK_RECORD");
    dw.dxfHex(5, 0x1B);
    //int psh = dw.handle();
    //dw.setPaperSpaceHandle(psh);
    //dw.dxfHex(330, 1);
    dw.dxfString(100, "AcDbSymbolTableRecord");
    dw.dxfString(100, "AcDbBlockTableRecord");
    dw.dxfString(  2, "*Paper_Space");
    dw.dxfHex(340, 0x1E);

    dw.dxfString(  0, "BLOCK_RECORD");
    dw.dxfHex(5, 0x23);
    //int ps0h = dw.handle();
    //dw.setPaperSpace0Handle(ps0h);
    //dw.dxfHex(330, 1);
    dw.dxfString(100, "AcDbSymbolTableRecord");
    dw.dxfString(100, "AcDbBlockTableRecord");
    dw.dxfString(  2, "*Paper_Space0");
    dw.dxfHex(340, 0x26);

    //dw.dxfString(  0, "ENDTAB");
}



/**
 * Writes a objects section. This section is needed in VER_R13.
 * Note that this method currently only writes a faked OBJECTS section
 * to make the file readable by Aut*cad.
 */
void DL_Dxf::writeObjects(DL_WriterA& dw) {
    //int dicId, dicId2, dicId3, dicId4, dicId5;
    //int dicId5;

    dw.dxfString(  0, "SECTION");
    dw.dxfString(  2, "OBJECTS");
    dw.dxfString(  0, "DICTIONARY");
    dw.dxfHex(5, 0xC);                            // C
    //dw.dxfHex(330, 0);
    dw.dxfString(100, "AcDbDictionary");
    dw.dxfInt(280, 0);
    dw.dxfInt(281, 1);
    dw.dxfString(  3, "ACAD_GROUP");
    //dw.dxfHex(350, dw.getNextHandle());          // D
    dw.dxfHex(350, 0xD);          // D
    dw.dxfString(  3, "ACAD_LAYOUT");
    dw.dxfHex(350, 0x1A);
    //dw.dxfHex(350, dw.getNextHandle()+0);        // 1A
    dw.dxfString(  3, "ACAD_MLINESTYLE");
    dw.dxfHex(350, 0x17);
    //dw.dxfHex(350, dw.getNextHandle()+1);        // 17
    dw.dxfString(  3, "ACAD_PLOTSETTINGS");
    dw.dxfHex(350, 0x19);
    //dw.dxfHex(350, dw.getNextHandle()+2);        // 19
    dw.dxfString(  3, "ACAD_PLOTSTYLENAME");
    dw.dxfHex(350, 0xE);
    //dw.dxfHex(350, dw.getNextHandle()+3);        // E
    dw.dxfString(  3, "AcDbVariableDictionary");
    dw.dxfHex(350, dw.getNextHandle());        // 2C
    dw.dxfString(  0, "DICTIONARY");
    dw.dxfHex(5, 0xD);
    //dw.handle();                                    // D
    //dw.dxfHex(330, 0xC);
    dw.dxfString(100, "AcDbDictionary");
    dw.dxfInt(280, 0);
    dw.dxfInt(281, 1);
    dw.dxfString(  0, "ACDBDICTIONARYWDFLT");
    dw.dxfHex(5, 0xE);
    //dicId4 = dw.handle();                           // E
    //dw.dxfHex(330, 0xC);                       // C
    dw.dxfString(100, "AcDbDictionary");
    dw.dxfInt(281, 1);
    dw.dxfString(  3, "Normal");
    dw.dxfHex(350, 0xF);
    //dw.dxfHex(350, dw.getNextHandle()+5);        // F
    dw.dxfString(100, "AcDbDictionaryWithDefault");
    dw.dxfHex(340, 0xF);
    //dw.dxfHex(340, dw.getNextHandle()+5);        // F
    dw.dxfString(  0, "ACDBPLACEHOLDER");
    dw.dxfHex(5, 0xF);
    //dw.handle();                                    // F
    //dw.dxfHex(330, dicId4);                      // E
    dw.dxfString(  0, "DICTIONARY");
    //dicId3 = dw.handle();                           // 17
    dw.dxfHex(5, 0x17);
    //dw.dxfHex(330, 0xC);                       // C
    dw.dxfString(100, "AcDbDictionary");
    dw.dxfInt(280, 0);
    dw.dxfInt(281, 1);
    dw.dxfString(  3, "Standard");
    dw.dxfHex(350, 0x18);
    //dw.dxfHex(350, dw.getNextHandle()+5);        // 18
    dw.dxfString(  0, "MLINESTYLE");
    dw.dxfHex(5, 0x18);
    //dw.handle();                                    // 18
    //dw.dxfHex(330, dicId3);                      // 17
    dw.dxfString(100, "AcDbMlineStyle");
    dw.dxfString(  2, "STANDARD");
    dw.dxfInt( 70, 0);
    dw.dxfString(  3, "");
    dw.dxfInt( 62, 256);
    dw.dxfReal( 51, 90.0);
    dw.dxfReal( 52, 90.0);
    dw.dxfInt( 71, 2);
    dw.dxfReal( 49, 0.5);
    dw.dxfInt( 62, 256);
    dw.dxfString(  6, "BYLAYER");
    dw.dxfReal( 49, -0.5);
    dw.dxfInt( 62, 256);
    dw.dxfString(  6, "BYLAYER");
    dw.dxfString(  0, "DICTIONARY");
    dw.dxfHex(5, 0x19);
    //dw.handle();                           // 17
    //dw.dxfHex(330, 0xC);                       // C
    dw.dxfString(100, "AcDbDictionary");
    dw.dxfInt(280, 0);
    dw.dxfInt(281, 1);
    dw.dxfString(  0, "DICTIONARY");
    //dicId2 = dw.handle();                           // 1A
    dw.dxfHex(5, 0x1A);
    //dw.dxfHex(330, 0xC);
    dw.dxfString(100, "AcDbDictionary");
    dw.dxfInt(281, 1);
    dw.dxfString(  3, "Layout1");
    dw.dxfHex(350, 0x1E);
    //dw.dxfHex(350, dw.getNextHandle()+2);        // 1E
    dw.dxfString(  3, "Layout2");
    dw.dxfHex(350, 0x26);
    //dw.dxfHex(350, dw.getNextHandle()+4);        // 26
    dw.dxfString(  3, "Model");
    dw.dxfHex(350, 0x22);
    //dw.dxfHex(350, dw.getNextHandle()+5);        // 22

    dw.dxfString(  0, "LAYOUT");
    dw.dxfHex(5, 0x1E);
    //dw.handle();                                    // 1E
    //dw.dxfHex(330, dicId2);                      // 1A
    dw.dxfString(100, "AcDbPlotSettings");
    dw.dxfString(  1, "");
    dw.dxfString(  2, "C:\\Program Files\\AutoCAD 2002\\plotters\\DWF ePlot (optimized for plotting).pc3");
    dw.dxfString(  4, "");
    dw.dxfString(  6, "");
    dw.dxfReal( 40, 0.0);
    dw.dxfReal( 41, 0.0);
    dw.dxfReal( 42, 0.0);
    dw.dxfReal( 43, 0.0);
    dw.dxfReal( 44, 0.0);
    dw.dxfReal( 45, 0.0);
    dw.dxfReal( 46, 0.0);
    dw.dxfReal( 47, 0.0);
    dw.dxfReal( 48, 0.0);
    dw.dxfReal( 49, 0.0);
    dw.dxfReal(140, 0.0);
    dw.dxfReal(141, 0.0);
    dw.dxfReal(142, 1.0);
    dw.dxfReal(143, 1.0);
    dw.dxfInt( 70, 688);
    dw.dxfInt( 72, 0);
    dw.dxfInt( 73, 0);
    dw.dxfInt( 74, 5);
    dw.dxfString(  7, "");
    dw.dxfInt( 75, 16);
    dw.dxfReal(147, 1.0);
    dw.dxfReal(148, 0.0);
    dw.dxfReal(149, 0.0);
    dw.dxfString(100, "AcDbLayout");
    dw.dxfString(  1, "Layout1");
    dw.dxfInt( 70, 1);
    dw.dxfInt( 71, 1);
    dw.dxfReal( 10, 0.0);
    dw.dxfReal( 20, 0.0);
    dw.dxfReal( 11, 420.0);
    dw.dxfReal( 21, 297.0);
    dw.dxfReal( 12, 0.0);
    dw.dxfReal( 22, 0.0);
    dw.dxfReal( 32, 0.0);
    dw.dxfReal( 14, 1.000000000000000E+20);
    dw.dxfReal( 24, 1.000000000000000E+20);
    dw.dxfReal( 34, 1.000000000000000E+20);
    dw.dxfReal( 15, -1.000000000000000E+20);
    dw.dxfReal( 25, -1.000000000000000E+20);
    dw.dxfReal( 35, -1.000000000000000E+20);
    dw.dxfReal(146, 0.0);
    dw.dxfReal( 13, 0.0);
    dw.dxfReal( 23, 0.0);
    dw.dxfReal( 33, 0.0);
    dw.dxfReal( 16, 1.0);
    dw.dxfReal( 26, 0.0);
    dw.dxfReal( 36, 0.0);
    dw.dxfReal( 17, 0.0);
    dw.dxfReal( 27, 1.0);
    dw.dxfReal( 37, 0.0);
    dw.dxfInt( 76, 0);
    //dw.dxfHex(330, dw.getPaperSpaceHandle());    // 1B
    dw.dxfHex(330, 0x1B);
    dw.dxfString(  0, "LAYOUT");
    dw.dxfHex(5, 0x22);
    //dw.handle();                                    // 22
    //dw.dxfHex(330, dicId2);                      // 1A
    dw.dxfString(100, "AcDbPlotSettings");
    dw.dxfString(  1, "");
    dw.dxfString(  2, "C:\\Program Files\\AutoCAD 2002\\plotters\\DWF ePlot (optimized for plotting).pc3");
    dw.dxfString(  4, "");
    dw.dxfString(  6, "");
    dw.dxfReal( 40, 0.0);
    dw.dxfReal( 41, 0.0);
    dw.dxfReal( 42, 0.0);
    dw.dxfReal( 43, 0.0);
    dw.dxfReal( 44, 0.0);
    dw.dxfReal( 45, 0.0);
    dw.dxfReal( 46, 0.0);
    dw.dxfReal( 47, 0.0);
    dw.dxfReal( 48, 0.0);
    dw.dxfReal( 49, 0.0);
    dw.dxfReal(140, 0.0);
    dw.dxfReal(141, 0.0);
    dw.dxfReal(142, 1.0);
    dw.dxfReal(143, 1.0);
    dw.dxfInt( 70, 1712);
    dw.dxfInt( 72, 0);
    dw.dxfInt( 73, 0);
    dw.dxfInt( 74, 0);
    dw.dxfString(  7, "");
    dw.dxfInt( 75, 0);
    dw.dxfReal(147, 1.0);
    dw.dxfReal(148, 0.0);
    dw.dxfReal(149, 0.0);
    dw.dxfString(100, "AcDbLayout");
    dw.dxfString(  1, "Model");
    dw.dxfInt( 70, 1);
    dw.dxfInt( 71, 0);
    dw.dxfReal( 10, 0.0);
    dw.dxfReal( 20, 0.0);
    dw.dxfReal( 11, 12.0);
    dw.dxfReal( 21, 9.0);
    dw.dxfReal( 12, 0.0);
    dw.dxfReal( 22, 0.0);
    dw.dxfReal( 32, 0.0);
    dw.dxfReal( 14, 0.0);
    dw.dxfReal( 24, 0.0);
    dw.dxfReal( 34, 0.0);
    dw.dxfReal( 15, 0.0);
    dw.dxfReal( 25, 0.0);
    dw.dxfReal( 35, 0.0);
    dw.dxfReal(146, 0.0);
    dw.dxfReal( 13, 0.0);
    dw.dxfReal( 23, 0.0);
    dw.dxfReal( 33, 0.0);
    dw.dxfReal( 16, 1.0);
    dw.dxfReal( 26, 0.0);
    dw.dxfReal( 36, 0.0);
    dw.dxfReal( 17, 0.0);
    dw.dxfReal( 27, 1.0);
    dw.dxfReal( 37, 0.0);
    dw.dxfInt( 76, 0);
    //dw.dxfHex(330, dw.getModelSpaceHandle());    // 1F
    dw.dxfHex(330, 0x1F);
    dw.dxfString(  0, "LAYOUT");
    //dw.handle();                                    // 26
    dw.dxfHex(5, 0x26);
    //dw.dxfHex(330, dicId2);                      // 1A
    dw.dxfString(100, "AcDbPlotSettings");
    dw.dxfString(  1, "");
    dw.dxfString(  2, "C:\\Program Files\\AutoCAD 2002\\plotters\\DWF ePlot (optimized for plotting).pc3");
    dw.dxfString(  4, "");
    dw.dxfString(  6, "");
    dw.dxfReal( 40, 0.0);
    dw.dxfReal( 41, 0.0);
    dw.dxfReal( 42, 0.0);
    dw.dxfReal( 43, 0.0);
    dw.dxfReal( 44, 0.0);
    dw.dxfReal( 45, 0.0);
    dw.dxfReal( 46, 0.0);
    dw.dxfReal( 47, 0.0);
    dw.dxfReal( 48, 0.0);
    dw.dxfReal( 49, 0.0);
    dw.dxfReal(140, 0.0);
    dw.dxfReal(141, 0.0);
    dw.dxfReal(142, 1.0);
    dw.dxfReal(143, 1.0);
    dw.dxfInt( 70, 688);
    dw.dxfInt( 72, 0);
    dw.dxfInt( 73, 0);
    dw.dxfInt( 74, 5);
    dw.dxfString(  7, "");
    dw.dxfInt( 75, 16);
    dw.dxfReal(147, 1.0);
    dw.dxfReal(148, 0.0);
    dw.dxfReal(149, 0.0);
    dw.dxfString(100, "AcDbLayout");
    dw.dxfString(  1, "Layout2");
    dw.dxfInt( 70, 1);
    dw.dxfInt( 71, 2);
    dw.dxfReal( 10, 0.0);
    dw.dxfReal( 20, 0.0);
    dw.dxfReal( 11, 12.0);
    dw.dxfReal( 21, 9.0);
    dw.dxfReal( 12, 0.0);
    dw.dxfReal( 22, 0.0);
    dw.dxfReal( 32, 0.0);
    dw.dxfReal( 14, 0.0);
    dw.dxfReal( 24, 0.0);
    dw.dxfReal( 34, 0.0);
    dw.dxfReal( 15, 0.0);
    dw.dxfReal( 25, 0.0);
    dw.dxfReal( 35, 0.0);
    dw.dxfReal(146, 0.0);
    dw.dxfReal( 13, 0.0);
    dw.dxfReal( 23, 0.0);
    dw.dxfReal( 33, 0.0);
    dw.dxfReal( 16, 1.0);
    dw.dxfReal( 26, 0.0);
    dw.dxfReal( 36, 0.0);
    dw.dxfReal( 17, 0.0);
    dw.dxfReal( 27, 1.0);
    dw.dxfReal( 37, 0.0);
    dw.dxfInt( 76, 0);
    //dw.dxfHex(330, dw.getPaperSpace0Handle());   // 23
    dw.dxfHex(330, 0x23);
    dw.dxfString(  0, "DICTIONARY");
    //dw.dxfHex(5, 0x2C);
    //dicId5 =
    dw.handle();                           // 2C
    //dw.dxfHex(330, 0xC);                       // C
    dw.dxfString(100, "AcDbDictionary");
    dw.dxfInt(281, 1);
    dw.dxfString(  3, "DIMASSOC");
    //dw.dxfHex(350, 0x2F);
    dw.dxfHex(350, dw.getNextHandle()+1);        // 2E
    dw.dxfString(  3, "HIDETEXT");
    //dw.dxfHex(350, 0x2E);
    dw.dxfHex(350, dw.getNextHandle());        // 2D
    dw.dxfString(  0, "DICTIONARYVAR");
    //dw.dxfHex(5, 0x2E);
    dw.handle();                                    // 2E
    //dw.dxfHex(330, dicId5);                      // 2C
    dw.dxfString(100, "DictionaryVariables");
    dw.dxfInt(280, 0);
    dw.dxfInt(  1, 2);
    dw.dxfString(  0, "DICTIONARYVAR");
    //dw.dxfHex(5, 0x2D);
    dw.handle();                                    // 2D
    //dw.dxfHex(330, dicId5);                      // 2C
    dw.dxfString(100, "DictionaryVariables");
    dw.dxfInt(280, 0);
    dw.dxfInt(  1, 1);
}


/**
 * Writes the end of the objects section. This section is needed in VER_R13.
 * Note that this method currently only writes a faked OBJECTS section
 * to make the file readable by Aut*cad.
 */
void DL_Dxf::writeObjectsEnd(DL_WriterA& dw) {
    dw.dxfString(  0, "ENDSEC");
}


